from .maxmincov import MaxMinCov
from .maxminbal import MaxMinBal
from ..core import ClusterData
from ..centers.boundedsepcenters import BoundedSepCenters
from ..distributions.gaussian import GaussianData
from ..distributions.exponential import ExpData
from ..distributions.standard_t import tData

class MaxMinClusters(ClusterData):
	"""
	"""

	def __init__(self, n_clusters, n_dim, n_samples, imbal_maxmin,
				 aspect_maxmin, radius_maxmin, min_sep=1, max_sep=1.5, 
				 aspect_ref=1.5, scale=1.0, packing=0.1,dist='gaussian'):

		cov_geom = MaxMinCov(ref_aspect=aspect_ref, aspect_maxmin=aspect_maxmin, 
							 radius_maxmin=radius_maxmin)
		center_geom = BoundedSepCenters(min_sep=min_sep,max_sep=max_sep, 
										packing=packing)
		class_bal = MaxMinBal(imbal_ratio=imbal_maxmin)

		if dist=='t':
			data_dist = tData()
		elif dist=='exp':
			data_dist = ExpData()
		else:
			data_dist = GaussianData()

		super().__init__(n_clusters,n_dim,n_samples,class_bal,cov_geom,
						 center_geom,data_dist,scale)




import numpy as np
from ..core import ClassBal
from .maxmin_sampler import maxmin_sampler

class MaxMinBal(ClassBal):
	"""
	Specify class imbalance by setting the ratio between the highest and lowest class sizes.
	"""

	def __init__(self, imbal_ratio):
		"""
		Creates a MaxMinBal object.
		"""

		self.imbal_ratio = imbal_ratio

	def float_to_int(self, float_class_sz, n_samples):
		class_sz = 1 + np.sort(np.round(float_class_sz))
		to_shrink = len(class_sz) - 1
		while (np.sum(class_sz) > n_samples):
			if (class_sz[to_shrink] > 1):
				class_sz[to_shrink] -= 1
				to_shrink -= 1
			else:
				to_shrink = len(class_sz) - 1
		return class_sz.astype(int)

	def make_class_sizes(self, clusterdata):
		n_samples = clusterdata.n_samples
		n_clusters = clusterdata.n_clusters

		# Set average class size as the reference size.
		ref_class_sz = n_samples/n_clusters

		# Determine minimum class size by requiring the average of the minimum and maximum
		# class sizes to be the reference size.
		min_class_sz = 2*ref_class_sz/(1 + self.imbal_ratio)

		# Set a pairwise sampling constraint to ensure that the sample sizes add to n_samples.
		f = lambda s: (2*ref_class_sz - s)

		# compute float class size estimates
		float_class_sz = maxmin_sampler(n_clusters, ref_class_sz, min_class_sz, self.imbal_ratio, f)

		# transform float class size estimates into integer class sizes
		class_sz = self.float_to_int(float_class_sz, n_samples)

		return class_sz








import numpy as np
import scipy.stats as stats
from ..core import CenterGeom
from scipy.spatial.distance import mahalanobis
from scipy.stats import chi2

class BoundedSepCenters(CenterGeom):
	"""
	Sample cluster centers by specifying maximum and minimum separation between centers.

	Cluster centers are generated using rejection sampling. Separation is computed with the
	Mahalanobis distance and is expressed in units of four times the standard deviation. 
	As a 1D example, a separation of 1.0 corresponds to two Gaussians that overlap at their
	2.5% quantiles.

	Parameters
	----------

	Attributes
	----------
	min_sep : float
		Minimum separation between clusters, in units of four times the standard deviation.
	max_sep : float
		Maximum separation between clusters, in units of four times the standard deviation.
	packing : float, default=0.1
		The ratio of total cluster volume over sampling volume. Cluster volume is computed
		from the maximum standard deviation among cluster shapes.
	"""
	
	def __init__(self, min_sep, max_sep=10, packing=.1):
		"""
		Creates a BoundedSepCenters object.
		"""
		self.min_sep = min_sep
		self.max_sep = max_sep
		self.packing = packing
		

	def place_centers(self, clusterdata):
		"""
		Place cluster centers sequentially with rejection sampling.

		Sample uniformly within a box and only accept new centers whose
		separation to all other clusters is greater than self.min_sep and 
		whose separation to at least one other cluster is less than 
		self.max_sep.
		"""

		# find the maximum sd
		cluster_sd = clusterdata.cluster_sd
		#max_sd = np.mean(np.array([np.mean(sd_vec) for sd_vec in cluster_sd]))
		max_sd = np.max(np.array([np.max(sd_vec) for sd_vec in cluster_sd]))

		cov_inv = clusterdata.cov_inv
		
		n_clusters = clusterdata.n_clusters
		n_dim = clusterdata.n_dim
		centers = np.zeros(shape=(n_clusters, n_dim))

		total_cluster_vol = n_clusters * ((4*max_sd)**n_dim)
		sampling_vol = total_cluster_vol / self.packing
		sampling_width = sampling_vol**(1/n_dim)

		# compute reference chi square value (take 0.95 quantile)
		chi_sq_ref = chi2.ppf(0.95, df=n_dim)

		# what's the reference distribution for the sum of squared scaled exponentials?
		# compute 0.95 quantile of distribution of the sum of n_dim Exp(1)**2 random variables

		# sum of squared scaled standard t distributions?
		
		for new_center in range(n_clusters):
			accept = False
			while not accept:
				proposed_center = sampling_width * np.random.uniform(size=n_dim)
				far_enough = True # Need to uphold minimum distance to other centers.
				close_enough = False # New center shouldn't be far away from all centers.

				# Check distances to previously selected centers
				for prev_ctr in range(new_center):
					chi_1 = mahalanobis(proposed_center, centers[prev_ctr], 
										 cov_inv[new_center]) ** 2
					chi_2 = mahalanobis(proposed_center, centers[prev_ctr], 
										 cov_inv[prev_ctr]) ** 2
					if (np.sqrt(np.min([chi_1, chi_2])) <= 2*np.sqrt(chi_sq_ref)*self.min_sep):
						far_enough = False 
						break

					if (np.sqrt(np.min([chi_1, chi_2])) <= 2*np.sqrt(chi_sq_ref)*self.max_sep):
						close_enough = True
				
				if ((far_enough and close_enough) or (new_center == 0)):
					accept = True
					#UNCOMMENT FOR DEBUGGING PURPOSES:
					#if not (new_center == 0): 
					#	print(chi_1, chi_2)
					centers[new_center,:] = proposed_center
					print('\t' + str(1+new_center) + '/' + str(n_clusters) + ' placed!')
				
		return(centers)


import numpy as np

def maxmin_sampler(n_samples, ref, min_val, maxmin_ratio, f_constrain):
	"""
	Generates samples around a reference value, with a fixed ratio between the maximum
	and minimum sample. Samples pairwise to enforce a further constraint on the samples.
	For example, the geometric mean of the samples can be specified.

	Parameters
	----------
	n_samples : int
	ref : float
	min_val : float
	maxmin_ratio : float
	f_constrain : function

	Returns
	-------
	out : ndarray

	"""

	if (maxmin_ratio == 1) or (min_val == 0):
		out = np.full(n_samples, fill_value=ref)
		return out

	max_val = min_val * maxmin_ratio
	
	if (n_samples > 2):
		# Besides min_val and max_val, only need n-2 samples
		n_gotta_sample = n_samples-2 
		samples = np.full(n_gotta_sample, fill_value=float(ref))
		# Sample according to triangular distribution with endpoints given by min_val
		# and max_val, and mode given by ref. Sample pairwise. The first sample in each
		# pair is generated randomly, and the second sample is calculated from the first.
		while (n_gotta_sample >= 2):
			samples[n_gotta_sample-1] = np.random.triangular(left=min_val, mode=ref, 
																right=max_val)
			samples[n_gotta_sample-2] = f_constrain(samples[n_gotta_sample-1])
			n_gotta_sample -= 2
		out = np.concatenate([[min_val], np.sort(samples), [max_val]])
	elif (n_samples == 2):
		out = np.array([min_val, max_val])
	elif (n_samples == 1):
		out = np.array([ref])
	return out


import numpy as np
import scipy.stats as stats
from .maxmin_sampler import maxmin_sampler
from ..core import CovGeom

class MaxMinCov(CovGeom):
	"""
	Constructs cluster covariance structures by setting ratios between maximum
	and minimum values of geometric parameters.


	Parameters
	----------
	ref_aspect : float			
		Reference aspect ratio for each cluster.
	aspect_maxmin : float
		Desired ratio between maximum and minimum aspect ratios among clusters.
	radius_max_min : float
		Desired ratio between maximum and minimum cluster radius.

	Attributes
	----------
	ref_aspect : float
	aspect_maxmin : float
	radius_maxmin : float
	"""
	
	def __init__(self, ref_aspect, aspect_maxmin, radius_maxmin):
		"""
		Constructs a MaxMinCov object.

		"""
		self.ref_aspect = ref_aspect
		self.aspect_maxmin = aspect_maxmin
		self.radius_maxmin = radius_maxmin
	

	def make_cluster_aspects(self, n_clusters):
		"""
		Generates aspect ratios (ratio between standard deviations along longest and shortest
		axes) for all clusters.

		Parameters
		----------
		n_clusters : int
			The number of clusters.

		Returns
		-------
		out : ndarray
			The aspect ratios for each cluster.

		"""

		min_aspect = 1 + (self.ref_aspect-1)/np.sqrt(self.aspect_maxmin)
		f = lambda a: ((self.ref_aspect-1)**2)/a
		return 1+maxmin_sampler(n_clusters, self.ref_aspect-1, min_aspect-1, self.aspect_maxmin, f)

		
	def make_cluster_radii(self, n_clusters, ref_radius, n_dim):
		""" 
		Computes cluster radii through constrained random sampling.

		The radius of a cluster is the geometric mean of the standard deviations along 
		the principal axes. Cluster radii are sampled such that the arithmetic mean of
		cluster volumes (cluster radius to the n_dim power) equals the reference volume
		(ref_radius to the n_dim power). The minimum and maximum radii are chosen so that
		the arithmetic average of the corresponding volumes equals the reference volume.

		Parameters
		----------
		n_clusters : int
			The number of clusters.
		ref_radius : float
			The reference radius for all clusters.
		n_dim : int
			The dimensionality of the clusters.

		Returns
		-------
		out : ndarray
			The cluster radii.
		"""
		min_radius = (2*(ref_radius**n_dim)/(1 + self.radius_maxmin**n_dim))**(1/n_dim)
		f = lambda r: (2*(ref_radius**n_dim) - (r**n_dim))**(1/n_dim)
		return maxmin_sampler(n_clusters, ref_radius, min_radius, self.radius_maxmin, f)
	

	def make_axis_sd(self, n_axes, sd, aspect):
		"""
		Generates standard deviations for the principal axes of a single cluster.

		Parameters
		----------
		n_axes : int
			Number of principal axes of this cluster
		sd : float
			Overall standard deviation for this cluster
		aspect : float
			Desired ratio between maximum and minimum axis standard deviations

		Returns
		-------

		"""
		min_sd = sd/np.sqrt(aspect)
		f = lambda s: (sd**2)/s
		return maxmin_sampler(n_axes, sd, min_sd, aspect, f)
		

	def make_cov(self, clusterdata, output_inv=True):
		"""
		Compute covariance structure for each cluster.

		Parameters
		----------
		clusterdata : a ClusterData object
			Specifies the number of clusters and other parameters
		output_inv : bool, optional
			If false, return only covariance matrices

		Returns
		-------
		out : a tuple (axis, sd, cov, cov_inv), where cov and cov_inv are lists of ndarrays
			The lists cov and cov_inv contain the covariance and inverse covariance
			matrices for each cluster. Corresponding list indices refer to the same
			cluster. Alternatively, if output_inv=False then only the list cov is
			returned.
		"""

		axis = list()
		sd = list()
		cov = list()
		cov_inv = list()

		n_clusters = clusterdata.n_clusters
		n_dim = clusterdata.n_dim
		scale = clusterdata.scale
		
		cluster_radii = self.make_cluster_radii(n_clusters, scale, n_dim)
		cluster_aspects = self.make_cluster_aspects(n_clusters)
		
		for clust in range(n_clusters):
			# compute principal axes for cluster
			axes = self.make_orthonormal_axes(n_dim, n_dim)
			axis_sd = self.make_axis_sd(n_dim, cluster_radii[clust], cluster_aspects[clust])

			axis.append(axes)
			sd.append(axis_sd)

			# can potentially not construct cov, cov_inv here; and instead do it only upon request
			cov.append(np.transpose(axes) @ np.diag(axis_sd**2) @ axes)
			cov_inv.append(np.transpose(axes) @ np.diag(1/axis_sd**2) @ axes)

		out = (axis, sd, cov, cov_inv)
			
		return out


from ..core import DataDist
import numpy as np

class GaussianData(DataDist):
	"""
	Generate multivariate Gaussian data for ClusterData.
	"""

	def __init__(self):
		"""
		Create a GaussianData object.
		"""
		pass

	def sample_cluster(self,cluster_size, mean,axes,sd):
		"""
		"""
		cov = np.transpose(axes) @ np.diag(sd**2) @ axes
		X = np.random.multivariate_normal(mean=mean, 
									cov=cov, size=cluster_size)
		return X



import numpy as np
import scipy.stats as stats

class ClusterData:
	"""
	Generates data sets with clusters according to user specifications.

	Parameters
	----------
	n_clusters : int
		Number of clusters
	n_dim : int
		Dimensionality of data points
	n_sample : int
		Total number of data points 
	class_bal : ClassBal object
		Defines relative class sizes (samples sizes for each cluster)
	cov_geom : CovGeom object
		Defines cluster covariance structures
	center_geom : CenterGeom object
		Defines placement of cluster centers
	data_dist : DataDist object
		Defines probability distribution for data
	scale : float, optional
		Sets reference length scale for simulated data

	Attributes
	----------
	class_sizes :
	cluster_axes :
	cluster_sd :
	cov : 
	cov_inv :
	data :

	"""
		
	def __init__(self, n_clusters, n_dim, n_samples, class_bal, cov_geom, 
		center_geom, data_dist, scale=1.0):
		"""
		Constructs a ClusterData object.
		"""
		self.n_clusters = n_clusters
		self.n_dim = n_dim
		self.n_samples = n_samples
		self.class_bal = class_bal
		self.cov_geom = cov_geom
		self.center_geom = center_geom
		self.data_dist = data_dist
		self.data = None
		self.labels = None
		self.centers = None
		self.class_sizes = None
		self.cov = None
		self.cov_inv = None
		self.cluster_axes = None
		self.cluster_sd = None
		self.scale = scale

	def to_csv(self, filename):
		"""
		Writes data to a csv file.
		"""
		if (self.data is None):
			print('Error: No data has been generated. Call ClusterData.generate_data to generate data.')
		else:
			np.savetxt(filename, self.data, delimiter=',')
		
		
	def generate_data(self,sparse=False, noise_factor=1.0, mode='sparse'):
		"""
		Generates a data set with clusters according to a ClusterData object.
		
		Parameters
		----------
		self : ClusterData
			Defines how data is generated
		sparse : bool
			Decide whether to add noise covariates to the data
		noise_factor : float
			Determines the number of noise covariates added, if sparse=True 
		mode : str
			Determines the meaning of noise_factor


		Returns
		-------
		X : (self.n_samples, 1+self.n_dim) ndarray
			Data matrix whose rows are the data points. The last column stores the 
			cluster labels as integers from 0 to self.n_clusters.
		"""
		print('Compute class sizes...')
		self.class_sizes = self.class_bal.make_class_sizes(self)

		print('Compute covariance structures...')
		axes, sd, cov, cov_inv = self.cov_geom.make_cov(self)
		self.cluster_axes = axes
		self.cluster_sd = sd
		self.cov = cov
		self.cov_inv = cov_inv

		print('Place cluster centers...')
		self.centers = self.center_geom.place_centers(self)

		print('Sample data...')
		self.data, self.labels = self.data_dist.sample(self)

		print('Success!')

		if sparse:
			self.data = ClusterData.add_noise(self.data,noise_factor=noise_factor,mode=mode)

		return (self.data, self.labels)


	def add_noise(self, noise_factor, mode='sparse', model='gaussian'):
		"""
		Add noise covariates to data set.

		If mode='sparse', noise_factor is the ratio between the number
		of noise covariates vs meaningful covariates. If mode='dim',
		noise_factor is the ratio between the total number of covariates
		and the number of samples, i.e. a high noise_factor indicates
		high-dimensional data in the statistical sense (more covariates 
		than samples).
		"""

		X = self.data
		mean_of_std = np.mean(np.array(self.cluster_sd))
		std_of_std = np.std(np.array(self.cluster_sd))

		n_dim = X.shape[1] 
		n_samples = X.shape[0]
		#all_obs = X.reshape((X.size,))
		#std_obs = np.std(all_obs); mean_obs = np.mean(all_obs)

		if (mode == 'sparse'):
			n_noise_cov = int(np.ceil(n_dim * noise_factor))
			synthetic_std = np.abs(np.random.normal(loc=mean_of_std, scale=std_of_std, size=n_noise_cov))
			noise_cov = np.random.normal(loc=0, scale=synthetic_std, size=(n_samples,n_noise_cov))
			return np.concatenate([X,noise_cov],axis=1)
		elif (mode == 'dim'):
			n_cov = int(np.ceil(n_samples * noise_factor))
			n_noise_cov = n_cov - n_dim
			if (n_noise_cov <= 0):
				print("Given data is already more high-dimensional than desired! No changes were made.")
				return X
			else:
				synthetic_std = np.abs(np.random.normal(loc=mean_of_std, scale=std_of_std, size=n_noise_cov))
				noise_cov = np.random.normal(loc=0, scale=synthetic_std, size=(n_samples,n_noise_cov))
				return np.concatenate([X,noise_cov],axis=1)



class CovGeom:
	"""
	Specifies the covariance structure for ClusterData.
	"""
	
	def __init__(self, clusterdata):
		pass

	def make_cov(self):
		pass

	def make_orthonormal_axes(self, n_dim, n_axes):
		"""
		Samples orthonormal axes for cluster generation.
		Creates n_axes orthonormal axes in n_dim-dimensional space.

		Parameters
		----------
		n_dim : int
			Dimensionality of the data space
		n_axes : int
			Number of axes to generate

		Returns
		-------
		out : (n_axes, n_dim) ndarray
			Each row is an axis.
		"""
		ortho_matrix = stats.ortho_group.rvs(n_dim)
		return ortho_matrix[:n_axes, :]



class CenterGeom:
	"""
	Defines placement of cluster centers for ClusterData.
	"""

	def __init__(self, clusterdata):
		pass

	def make_centers(self):
		pass



class ClassBal:
	"""
	Specifies the relative cluster sizes for ClusterData.
	"""

	def __init__(self):
		pass

	def make_class_sizes(self, clusterdata):
		pass



class DataDist:
	"""
	Defines data probability distribution for ClusterData.
	"""

	def __init__(self):
		pass

	def sample(self, clusterdata):
		n_clusters = clusterdata.n_clusters
		n_samples = clusterdata.n_samples
		n_dim = clusterdata.n_dim
		class_sizes = clusterdata.class_sizes
		centers = clusterdata.centers

		axes = clusterdata.cluster_axes
		sd = clusterdata.cluster_sd

		X = np.full(shape=(n_samples, n_dim), fill_value=np.nan)
		y = np.full(n_samples, fill_value=np.nan).astype(int)

		start = 0
		for i in range(n_clusters):
			end = start + class_sizes[i]
			# Set class label
			y[start:end] = i
			# Sample data
			X[start:end,:] = self.sample_cluster(cluster_size=class_sizes[i], 
										  mean=centers[i], axes=axes[i],
										  sd=sd[i])
			start = end

		return (X, y)

	def sample_cluster(self,cluster_size, mean,axes,sd):
		pass


from ..core import DataDist
import numpy as np

class ExpData(DataDist):
	"""
	Generates multivariate exponential data for ClusterData.
	"""

	def __init__(self):
		"""
		Create a ExpData object.
		"""
		pass

	def sample_cluster(self,cluster_size, mean,axes,sd):
		"""
		cluster_size : int
		mean : ndarray
		axes : list of ndarray
		sd : list of float
		"""

		# each row of axes is an axis
		n_axes = axes.shape[0]

		X = np.zeros(shape=(cluster_size,n_axes));

		# sample along each principal axis independently
		for i, stdev in enumerate(sd):
			# use exponential with desired standard deviation and random sign
			signs = np.random.choice(a=[-1,+1],size=cluster_size,replace=True)
			X[:,i] = np.random.exponential(scale=stdev, size=cluster_size) * signs

		# transform X from principal axis coordinates to standard coordinates
		X = X @ axes

		# mean-shift X
		X = X + mean[np.newaxis,:]

		return X


from ..core import DataDist
import numpy as np

class tData(DataDist):
	"""
	Generates multivariate t-distributed data for ClusterData.
	"""

	def __init__(self, df=1):
		"""
		Create a tData object.
		"""
		self.df = df

	def sample_cluster(self,cluster_size, mean,axes,sd):
		"""
		cluster_size : int
		mean : ndarray
		axes : list of ndarray
		sd : list of float
		"""

		# compute median for absolute value of t distribution
		n = 1000000
		abs_t_median = np.median(np.random.standard_t(df=self.df,size=n))

		# each row of axes is an axis
		n_axes = axes.shape[0]
		X = np.zeros(shape=(cluster_size,n_axes));

		# sample along each principal axis independently
		for i, stdev in enumerate(sd):
			# scale t distribution so that its median equals stdev
			scaling = stdev/(2*abs_t_median)
			X[:,i] = scaling * np.random.standard_t(df=self.df, size=cluster_size)

		# transform X from principal axis coordinates to standard coordinates
		X = X @ axes

		# mean-shift X
		X = X + mean[np.newaxis,:]

		return X